// () ..,2003
// http://alwd.narod.ru
// mailto:alwd@narod.ru

unit stacks;

interface

uses SysUtils;

const STACK_MINSTACKDEPTH = 2;
const STACK_MAXSTACKDEPTH = 1024;

type

  //         :
  //      .     
  //         .
  //        Delphi 
  //     .

  EStackError = class(Exception);


  //    .     
  //          .
  //         
  //   fPrimeElem.
  //        , 
  //       UndoAll,
  //    PopPrime.   ,  
  //     (  ""),
  //    X      .
  //     PopPrime ,    
  //  .
  //    ,   Stack.Clear .
  //     Clear (     Stack)
  //   PopPrime  .

  TVariantStack = class
    private
      fValuePointers: array[0..STACK_MAXSTACKDEPTH] of ^variant;
      fElementsCount: word;
      fMaxDepth: word;
      fTop: word;
      fIsClean: boolean;
      fPrimeElem: variant;
      procedure IncTop;
      procedure DecTop;
      procedure CheckDepth(var Depth: word);
      procedure SetDepth(Depth: word); virtual;
      //    : (Delphi 3)
      //  FreeMem,  ,     nil.
      procedure FreeElem(i: word);
    public
      constructor Create(Depth: word);
      destructor Destroy; override;
    //   
      property MaxDepth: word read fMaxDepth write SetDepth;
    //    
      procedure Push(x: variant);
    //    
      procedure Pop(var x: variant); 
    //       
      procedure PopOlder(var x: variant);
    //  ,      
    //    Push(x)
      procedure PopPrime(var x: variant);
    //    ,    
    //       Push(x)
      procedure Clear;
    //  ,    Push
    //      
      function IsFull: boolean;
    //  
      function IsEmpty: boolean;
    //      
      function FreeCellsCount: word;
    //        
      property IsClean: boolean read fIsClean;
    //    
      property ElementsCount: word read fElementsCount;
  end;


  //   TVariantSTack    
  //   .

  TAnyValueStack = class(TVariantStack)
    private
      fValuePointers: array[0..STACK_MAXSTACKDEPTH] of pointer;
      fPrimeElem: pointer;
      //      
      procedure PopToNull;
      procedure SetDepth(Depth: word); override;
      //   ,      ;
      //        TVariantStack.FreeElem,
      //        .
      procedure FreeElem(i: word);
    public
      constructor Create(Depth: word);
      destructor Destroy; override;
    //    
      procedure Push(var x; Size: word);
    //    
      procedure Pop(var x; Size: word);
    //       
      procedure PopOlder(var x; Size: word);
    //  ,      
    //    Push(x)
      procedure PopPrime(var x; Size: word);
    //    ,    
    //       Push(x, Size)
      procedure Clear;
  end;


implementation


(*==================== TVariantStack ==========================*)

constructor TVariantStack.Create(Depth: word);
var
  i: word;
begin

  CheckDepth(Depth);
  fMaxDepth := Depth;
  //         
  fElementsCount := 0;
  //        .
  fTop := 0;
  //       .
  for i := 0 to STACK_MAXSTACKDEPTH do fValuePointers[i] := nil;
  // ,       .
  fIsClean := TRUE;

  inherited Create;

end;


destructor TVariantStack.Destroy;
var
  x: variant;
begin
  //      ,
  while not IsEmpty do Pop(x);
  //     .
  inherited Destroy;
end;


procedure TVariantStack.CheckDepth;
begin
  if Depth < STACK_MINSTACKDEPTH then Depth := STACK_MINSTACKDEPTH;
  if Depth > STACK_MAXSTACKDEPTH then Depth := STACK_MAXSTACKDEPTH;
end;


procedure TVariantStack.FreeElem;
begin
  if fValuePointers[i] <> nil then begin
    FreeMem(fValuePointers[i]);
    fValuePointers[i] := nil;
  end;
end;


procedure TVariantStack.IncTop;
begin
  if fTop < (fMaxDepth - 1) then Inc(fTop) else fTop := 0;
end;


procedure TVariantStack.DecTop;
begin
  if fTop < 1 then fTop := fMaxDepth - 1 else Dec(fTop);
end;


procedure TVariantStack.Push;
begin

  //      ,    
  //          "" ;
  //       ,
  //  ..  variant      ,
  //    ,       .
  if fElementsCount >= fMaxDepth then begin
    FreeElem(fTop);
    //     ,    
    //       .
    Dec(fElementsCount);
  end;

  //      ,
  GetMem(fValuePointers[fTop],SizeOf(x));
  //     ,
  fValuePointers[fTop]^ := x;
  //    
  Inc(fElementsCount);
  //        (..   )
  IncTop;
  //       PopPrime.
  //   ,   .
  if fIsClean then begin
    fPrimeElem := x;
    fIsClean := FALSE;
  end;

end;


procedure TVariantStack.Pop;
begin

  //       ,  .
  //   ,  .
  if IsEmpty then begin
    if fIsClean then raise EStackError.Create('TVariantStack ERROR: Pop(x) called under clean stack')
    else raise EStackError.Create('TVariantStack ERROR: Pop(x) called under empty stack');
    end
  else begin
    //    ,
    Dec(fElementsCount);
    //      ,
    DecTop;
    //    
    x := fValuePointers[fTop]^;
    //   
    FreeElem(fTop);
  end;

end;


procedure TVariantStack.PopOlder;
begin
  //       ,  .
  //   ,  .
  if IsEmpty then begin
    if fIsClean then raise EStackError.Create('TVariantStack ERROR: PopOlder(x) called under clean stack')
    else raise EStackError.Create('TVariantStack ERROR: PopOlder(x) called under empty stack');
    end
  else while not IsEmpty do Pop(x);
end;


procedure TVariantStack.PopPrime;
begin
  if fIsClean then begin
    raise EStackError.Create('TVariantStack ERROR: PopPrime is down for clean stack');
    end
  else begin
    while not IsEmpty do Pop(x);
    x := fPrimeElem;
  end;
end;


procedure TVariantStack.Clear;
var
  x: variant;
begin
  //       
  while not IsEmpty do Pop(x);
  fIsClean := TRUE;
end;


function TVariantStack.IsFull;
begin
  Result := (fElementsCount = fMaxDepth);
end;


function TVariantStack.IsEmpty;
begin
  Result := (fElementsCount = 0);
end;


function TVariantStack.FreeCellsCount;
begin
  Result := (fMaxDepth - fElementsCount);
end;


{$WARNINGS OFF}
procedure TVariantStack.SetDepth(Depth: word);
var
  i: word;
  Count: word;
  ElPresent: boolean;
  Elements: array[0..STACK_MAXSTACKDEPTH] of variant;
begin

  CheckDepth(Depth);

  //      ,
  //  ..       
  //        .   
  //       
  //  ,     .

  //        ,
  ElPresent := not IsEmpty;
  if ElPresent then begin
    Count := fElementsCount - 1;
    for i := Count downto 0 do Pop(Elements[i]);
  end;
  //     ,
  fMaxDepth := Depth;
  //     ,       
  fTop := 0;
  //       
  if ElPresent then for i := 0 to Count do Push(Elements[i]);

end;
{$WARNINGS ON}

(*===================== End of TVariantStack =====================*)


(*====================== TAnyValueStack ==========================*)

// Private functions of TAnyValueStack

procedure TAnyValueStack.PopToNull;
begin
  Dec(fElementsCount);
  DecTop;
  FreeElem(fTop);
end;


procedure TAnyValueStack.FreeElem;
begin
  if fValuePointers[i] <> nil then begin
    FreeMem(fValuePointers[i]);
    fValuePointers[i] := nil;
  end;
end;


procedure TAnyValueStack.SetDepth(Depth: word);
var
  i: integer;
  Count: word;
  SurplusCount: integer;
  Elements: array[0..STACK_MAXSTACKDEPTH] of pointer;
begin

  CheckDepth(Depth);

  //      ,
  //  ..       
  //        .   
  //       
  //  ,     .

  if not IsEmpty then begin
    //      .
    Count := fElementsCount - 1;
    //         
    for i := Count downto 0 do begin
      DecTop;
      Elements[i] := fValuePointers[fTop];
    end;
    //     ,
    fTop := 0;
    //        ,
    //   .
    //  ,    
    if Depth < fElementsCount then SurplusCount := fElementsCount - Depth
    else SurplusCount := 0;
    // ,    ,  ;
    for i := 0 to SurplusCount - 1 do FreeMem(Elements[i]);
    //       
    //     
    fMaxDepth := Depth;
    for i := SurplusCount to Count do begin
      fValuePointers[fTop] := Elements[i];
      IncTop;
    end;
    fElementsCount := fElementsCount - SurplusCount;
  end;

  //   ,      IF  ,
  //     .
  fMaxDepth := Depth;

end;


// Public functions of TAnyValueStack

constructor TAnyValueStack.Create;
begin
  fPrimeElem := nil;
  inherited Create(Depth);
end;


destructor TAnyValueStack.Destroy;
begin
  while not IsEmpty do PopToNull;
  if not fIsClean then FreeMem(fPrimeElem);
  inherited Destroy;
end;


procedure TAnyValueStack.Push;
begin

  //      ,    
  //          "" ;
  //     .
  if fElementsCount >= fMaxDepth then begin
    FreeElem(fTop);
    //     ,    
    //       .
    Dec(fElementsCount);
  end;

  //      ,
  GetMem(Self.fValuePointers[Self.fTop],Size);
  //     ,
  Move(x,fValuePointers[fTop]^,Size);
  //    
  Inc(fElementsCount);
  //        (..   )
  IncTop;
  //       PopPrime.
  //   ,   .
  if fIsClean then begin
    GetMem(fPrimeElem,Size);
    Move(x,fPrimeElem^,Size);
    fIsClean := FALSE;
  end;

end;


procedure TAnyValueStack.Pop;
begin

  //       ,  .
  //   ,  .
  if IsEmpty then begin
    if fIsClean then raise EStackError.Create('TAnyValueStack ERROR: Pop(x) called under clean stack')
    else raise EStackError.Create('TAnyValueStack ERROR: Pop(x) called under empty stack');
    end
  else begin
    //    ,
    Dec(fElementsCount);
    //      ,
    DecTop;
    //    
    Move(fValuePointers[fTop]^,x,Size);
    //   
    FreeElem(fTop);
  end;

end;


procedure TAnyValueStack.PopOlder;
begin
  if IsEmpty then begin
    if fIsClean then raise EStackError.Create('TAnyValueStack ERROR: PopOlder(x) called under clean stack')
    else raise EStackError.Create('TAnyValueStack ERROR: PopOlder(x) called under empty stack');
    end
  else begin
    //     
    while fElementsCount > 1 do PopToNull;
    //   
    Pop(x,Size);
  end;
end;


procedure TAnyValueStack.PopPrime;
begin
  if fIsClean then begin
    raise EStackError.Create('TAnyValueStack ERROR: PopPrime is down for clean stack');
    end
  else begin
    //  
    while not IsEmpty do PopToNull;
    //   
    Move(fPrimeElem^,x,Size);
  end;
end;


procedure TAnyValueStack.Clear;
begin
  //   ,   
  if fIsClean then Exit;
  //  
  while not IsEmpty do PopToNull;
  //   
  FreeMem(fPrimeElem);
  fPrimeElem := nil;
  // ,   
  fIsClean := TRUE;
end;


end.
